## contains the entire YogaSutra as a list of strings. 
## used to test the functions in the process_sanskrit module

ys = [
 'atha yogānuśāsanam ',
 'yogaś cittavṛttinirodhaḥ ',
 "tadā draṣṭuḥ svarūpe 'vasthānam ",
 'vṛttisārūpyam itaratra ',
 'vṛttayaḥ pañcatayyaḥ kliṣṭākliṣṭāḥ ',
 'pramāṇaviparyayavikalpanidrāsmṛtayaḥ ',
 'pratyakṣānumānāgamāḥ pramāṇāni ',
 'viparyayo mithyājñānam atadrūpapratiṣṭham ',
 'śabdajñānānupātī vastuśūnyo vikalpaḥ ',
 'abhāvapratyayālambanā vṛttir nidrā ',
 'anubhūtaviṣayāsaṃpramoṣaḥ smṛtiḥ ',
 'abhyāsavairāgyābhyāṃ tannirodhaḥ ',
 "tatra sthitau yatno 'bhyāsaḥ ",
 'sa tu dīrghakālanairantaryasatkārāsevito dṛḍhabhūmiḥ ',
 'dṛṣṭānuśravikaviṣayavitṛṣṇasya vaśīkārasaṃjñā vairāgyam ',
 'tat paraṃ puruṣakhyāter guṇavaitṛṣṇyam ',
 'vitarkavicārānandāsmitārūpānugamāt saṃprajñātaḥ ',
 "virāmapratyayābhyāsapūrvaḥ saṃskāraśeṣo 'nyaḥ ",
 'bhavapratyayo videhaprakṛtilayānām ',
 'śraddhāvīryasmṛtisamādhiprajñāpūrvaka itareṣām ',
 'tīvrasaṃvegānām āsannaḥ ',
 "mṛdumadhyādhimātratvāt tato 'pi viśeṣaḥ ",
 'īśvarapraṇidhānād vā ',
 'kleśakarmavipākāśayair aparāmṛṣṭaḥ puruṣaviśeṣa īśvaraḥ ',
 'tatra niratiśayaṃ sarvajñabījam ',
 'pūrveṣām api guruḥ kālenānavacchedāt ',
 'tasya vācakaḥ praṇavaḥ ',
 'tajjapas tadarthabhāvanam ',
 "tataḥ pratyakcetanādhigamo 'py antarāyābhāvaś ca ",
 "vyādhistyānasaṃśayapramādālasyāviratibhrāntidarśanālabdhabhūmikatvānavasthitatvāni cittavikṣepās te 'ntarāyāḥ ",
 'duḥkhadaurmanasyāṅgamejayatvaśvāsapraśvāsā vikṣepasahabhuvaḥ ',
 'tatpratiṣedhārtham ekatattvābhyāsaḥ ',
 'maitrīkaruṇāmuditopekṣāṇāṃ sukhaduḥkhapuṇyāpuṇyaviṣayāṇāṃ bhāvanātaś cittaprasādanam ',
 'pracchardanavidhāraṇābhyāṃ vā prāṇasya ',
 'viṣayavatī vā pravṛttir utpannā manasaḥ sthitinibandhanī ',
 'viśokā vā jyotiṣmatī ',
 'vītarāgaviṣayaṃ vā cittam ',
 'svapnanidrājñānālambanaṃ vā ',
 'yathābhimatadhyānād vā ',
 "paramāṇuparamamahattvānto 'sya vaśīkāraḥ ",
 'kṣīṇavṛtter abhijātasyeva maṇer grahītṛgrahaṇagrāhyeṣu tatsthatadañjanatā samāpattiḥ ',
 'tatra śabdārthajñānavikalpaiḥ saṃkīrṇā savitarkā samāpattiḥ ',
 'smṛtipariśuddhau svarūpaśūnyevārthamātranirbhāsā nirvitarkā ',
 'etayaiva savicārā nirvicārā ca sūkṣmaviṣayā vyākhyātā ',
 'sūkṣmaviṣayatvaṃ cāliṅgaparyavasānam ',
 'tā eva sabījaḥ samādhiḥ ',
 "nirvicāravaiśāradye 'dhyātmaprasādaḥ ",
 'ṛtaṃbharā tatra prajñā ',
 'śrutānumānaprajñābhyām anyaviṣayā viśeṣārthatvāt ',
 "tajjaḥ saṃskāro 'nyasaṃskārapratibandhī ",
 'tasyāpi nirodhe sarvanirodhān nirbījaḥ samādhiḥ ',
 'tapaḥsvādhyāyeśvarapraṇidhānāni kriyāyogaḥ ',
 'samādhibhāvanārthaḥ kleśatanūkaraṇārthaś ca ',
 'avidyāsmitārāgadveṣābhiniveśāḥ kleśāḥ ',
 'avidyā kṣetram uttareṣāṃ prasuptatanuvicchinnodārāṇām ',
 'anityāśuciduḥkhānātmasu nityaśucisukhātmakhyātir avidyā ',
 'dṛgdarśanaśaktyor ekātmatevāsmitā ',
 'sukhānuśayī rāgaḥ ',
 'duḥkhānuśayī dveṣaḥ ',
 "svarasavāhī viduṣo 'pi tathā rūḍho 'bhiniveśaḥ ",
 'te pratiprasavaheyāḥ sūkṣmāḥ ',
 'dhyānaheyās tadvṛttayaḥ ',
 'kleśamūlaḥ karmāśayo dṛṣṭādṛṣṭajanmavedanīyaḥ ',
 'sati mūle tadvipāko jātyāyurbhogāḥ ',
 'te hlādaparitāpaphalāḥ puṇyāpuṇyahetutvāt ',
 'pariṇāmatāpasaṃskāraduḥkhair guṇavṛttivirodhāc ca duḥkham eva sarvaṃ vivekinaḥ ',
 'heyaṃ duḥkham anāgatam ',
 'draṣṭṛdṛśyayoḥ saṃyogo heyahetuḥ ',
 'prakāśakriyāsthitiśīlaṃ bhūtendriyātmakaṃ bhogāpavargārthaṃ dṛśyam ',
 'viśeṣāviśeṣaliṅgamātrāliṅgāni guṇaparvāṇi ',
 "draṣṭā dṛśimātraḥ śuddho 'pi pratyayānupaśyaḥ ",
 'tadartha eva dṛśyasyātmā ',
 'kṛtārthaṃ prati naṣṭam apy anaṣṭaṃ tadanyasādhāraṇatvāt ',
 'svasvāmiśaktyoḥ svarūpopalabdhihetuḥ saṃyogaḥ ',
 'tasya hetur avidyā ',
 'tadabhāvāt saṃyogābhāvo hānaṃ taddṛśeḥ kaivalyam ',
 'vivekakhyātir aviplavā hānopāyaḥ ',
 'tasya saptadhā prāntabhūmiḥ prajñā ',
 'yogāṅgānuṣṭhānād aśuddhikṣaye jñānadīptir ā vivekakhyāteḥ ',
 "yamaniyamāsanaprāṇāyāmapratyāhāradhāraṇādhyānasamādhayo 'ṣṭāv aṅgāni ",
 'ahiṃsāsatyāsteyabrahmacaryāparigrahā yamāḥ ',
 'jātideśakālasamayānavacchinnāḥ sārvabhaumā mahāvratam ',
 'śaucasaṃtoṣatapaḥsvādhyāyeśvarapraṇidhānāni niyamāḥ ',
 'vitarkabādhane pratipakṣabhāvanam ',
 'vitarkā hiṃsādayaḥ kṛtakāritānumoditā lobhakrodhamohapūrvakā mṛdumadhyādhimātrā duḥkhājñānānantaphalā iti pratipakṣabhāvanam ',
 'ahiṃsāpratiṣṭhāyāṃ tatsaṃnidhau vairatyāgaḥ ',
 'satyapratiṣṭhāyāṃ kriyāphalāśrayatvam ',
 'asteyapratiṣṭhāyāṃ sarvaratnopasthānam ',
 'brahmacaryapratiṣṭhāyāṃ vīryalābhaḥ ',
 'aparigrahasthairye janmakathaṃtāsaṃbodhaḥ ',
 'śaucāt svāṅgajugupsā parair asaṃsargaḥ ',
 'sattvaśuddhisaumanasyaikāgryendriyajayātmadarśanayogyatvāni ca ',
 'saṃtoṣād anuttamaḥ sukhalābhaḥ ',
 'kāyendriyasiddhir aśuddhikṣayāt tapasaḥ ',
 'svādhyāyād iṣṭadevatāsaṃprayogaḥ ',
 'samādhisiddhir īśvarapraṇidhānāt ',
 'sthirasukham āsanam ',
 'prayatnaśaithilyānantasamāpattibhyām ',
 'tato dvandvānabhighātaḥ ',
 'tasmin sati śvāsapraśvāsayor gativicchedaḥ prāṇāyāmaḥ ',
 'bāhyābhyantarastambhavṛttir deśakālasaṃkhyābhiḥ paridṛṣṭo dīrghasūkṣmaḥ ',
 'bāhyābhyantaraviṣayākṣepī caturthaḥ ',
 'tataḥ kṣīyate prakāśāvaraṇam ',
 'dhāraṇāsu ca yogyatā manasaḥ ',
 'svaviṣayāsaṃprayoge cittasvarūpānukāra ivendriyāṇāṃ pratyāhāraḥ ',
 'tataḥ paramā vaśyatendriyāṇām ',
 'deśabandhaś cittasya dhāraṇā ',
 'tatra pratyayaikatānatā dhyānam ',
 'tad evārthamātranirbhāsaṃ svarūpaśūnyam iva samādhiḥ ',
 'trayam ekatra saṃyamaḥ ',
 'tajjayāt prajñālokaḥ ',
 'tasya bhūmiṣu viniyogaḥ ',
 'trayam antaraṅgaṃ pūrvebhyaḥ ',
 'tad api bahiraṅgaṃ nirbījasya ',
 'vyutthānanirodhasaṃskārayor abhibhavaprādurbhāvau nirodhakṣaṇacittānvayo nirodhapariṇāmaḥ ',
 'tasya praśāntavāhitā saṃskārāt ',
 'sarvārthataikāgratayoḥ kṣayodayau cittasya samādhipariṇāmaḥ ',
 'tataḥ punaḥ śāntoditau tulyapratyayau cittasyaikāgratāpariṇāmaḥ ',
 'etena bhūtendriyeṣu dharmalakṣaṇāvasthāpariṇāmā vyākhyātāḥ ',
 'śāntoditāvyapadeśyadharmānupātī dharmī ',
 'kramānyatvaṃ pariṇāmānyatve hetuḥ ',
 'pariṇāmatrayasaṃyamād atītānāgatajñānam ',
 'śabdārthapratyayānām itaretarādhyāsāt saṃkaras tatpravibhāgasaṃyamāt sarvabhūtarutajñānam ',
 'saṃskārasākṣātkaraṇāt pūrvajātijñānam ',
 'pratyayasya paracittajñānam ',
 'na ca tat sālambanaṃ tasyāviṣayībhūtatvāt ',
 "kāyarūpasaṃyamāt tadgrāhyaśaktistambhe cakṣuḥprakāśāsaṃprayoge 'ntardhānam ",
 'sopakramaṃ nirupakramaṃ ca karma tatsaṃyamād aparāntajñānam ariṣṭebhyo vā ',
 'maitryādiṣu balāni ',
 'baleṣu hastibalādīni ',
 'pravṛttyālokanyāsāt sūkṣmavyavahitaviprakṛṣṭajñānam ',
 'bhuvanajñānaṃ sūrye saṃyamāt ',
 'candre tārāvyūhajñānam ',
 'dhruve tadgatijñānam ',
 'nābhicakre kāyavyūhajñānam ',
 'kaṇṭhakūpe kṣutpipāsānivṛttiḥ ',
 'kūrmanāḍyāṃ sthairyam ',
 'mūrdhajyotiṣi siddhadarśanam ',
 'prātibhād vā sarvam ',
 'hṛdaye cittasaṃvit ',
 'sattvapuruṣayor atyantāsaṃkīrṇayoḥ pratyayāviśeṣo bhogaḥ parārthāt svārthasaṃyamāt puruṣajñānam ',
 'tataḥ prātibhaśrāvaṇavedanādarśāsvādavārtā jāyante ',
 'te samādhāv upasargā vyutthāne siddhayaḥ ',
 'bandhakāraṇaśaithilyāt pracārasaṃvedanāc ca cittasya paraśarīrāveśaḥ ',
 'udānajayāj jalapaṅkakaṇṭakādiṣv asaṅga utkrāntiś ca ',
 'samānajayāj jvalanam ',
 'śrotrākāśayoḥ saṃbandhasaṃyamād divyaṃ śrotram ',
 'kāyākāśayoḥ saṃbandhasaṃyamāl laghutūlasamāpatteś cākāśagamanam ',
 'bahir akalpitā vṛttir mahāvidehā tataḥ prakāśāvaraṇakṣayaḥ ',
 'sthūlasvarūpasūkṣmānvayārthavattvasaṃyamād bhūtajayaḥ ',
 "tato 'ṇimādiprādurbhāvaḥ kāyasaṃpat taddharmānabhighātaś ca ",
 'rūpalāvaṇyabalavajrasaṃhananatvāni kāyasaṃpat ',
 'grahaṇasvarūpāsmitānvayārthavattvasaṃyamād indriyajayaḥ ',
 'tato manojavitvaṃ vikaraṇabhāvaḥ pradhānajayaś ca ',
 'sattvapuruṣānyatākhyātimātrasya sarvabhāvādhiṣṭhātṛtvaṃ sarvajñātṛtvaṃ ca ',
 'tadvairāgyād api doṣabījakṣaye kaivalyam ',
 'sthānyupanimantraṇe saṅgasmayākaraṇaṃ punar aniṣṭaprasaṅgāt ',
 'kṣaṇatatkramayoḥ saṃyamād vivekajaṃ jñānam ',
 'jātilakṣaṇadeśair anyatānavacchedāt tulyayos tataḥ pratipattiḥ ',
 'tārakaṃ sarvaviṣayaṃ sarvathāviṣayam akramaṃ ceti vivekajaṃ jñānam ',
 'sattvapuruṣayoḥ śuddhisāmye kaivalyam iti ',
 'janmauṣadhimantratapaḥsamādhijāḥ siddhayaḥ ',
 'jātyantarapariṇāmaḥ prakṛtyāpūrāt ',
 'nimittam aprayojakaṃ prakṛtīnāṃ varaṇabhedas tu tataḥ kṣetrikavat ',
 'nirmāṇacittāny asmitāmātrāt ',
 'pravṛttibhede prayojakaṃ cittam ekam anekeṣām ',
 'tatra dhyānajam anāśayam ',
 'karmāśuklākṛṣṇaṃ yoginas trividham itareṣām ',
 'tatas tadvipākānuguṇānām evābhivyaktir vāsanānām ',
 'jātideśakālavyavahitānām apy ānantaryaṃ smṛtisaṃskārayor ekarūpatvāt ',
 'tāsām anāditvaṃ cāśiṣo nityatvāt ',
 'hetuphalāśrayālambanaiḥ saṃgṛhītatvād eṣām abhāve tadabhāvaḥ ',
 "atītānāgataṃ svarūpato 'sty adhvabhedād dharmāṇām ",
 'te vyaktasūkṣmā guṇātmānaḥ ',
 'pariṇāmaikatvād vastutattvam ',
 'vastusāmye cittabhedāt tayor vibhaktaḥ panthāḥ ',
 'na caikacittatantraṃ vastu tadapramāṇakaṃ tadā kiṃ syāt ',
 'taduparāgāpekṣitvāc cittasya vastu jñātājñātam ',
 'sadā jñātāś cittavṛttayas tatprabhoḥ puruṣasyāpariṇāmitvāt ',
 'na tat svābhāsaṃ dṛśyatvāt ',
 'ekasamaye cobhayānavadhāraṇam ',
 'cittāntaradṛśye buddhibuddher atiprasaṅgaḥ smṛtisaṃkaraś ca ',
 'citer apratisaṃkramāyās tadākārāpattau svabuddhisaṃvedanam ',
 'draṣṭṛdṛśyoparaktaṃ cittaṃ sarvārtham ',
 'tad asaṃkhyeyavāsanābhiś citram api parārthaṃ saṃhatyakāritvāt ',
 'viśeṣadarśina ātmabhāvabhāvanānivṛttiḥ ',
 'tadā vivekanimnaṃ kaivalyaprāgbhāraṃ cittam ',
 'tacchidreṣu pratyayāntarāṇi saṃskārebhyaḥ ',
 'hānam eṣāṃ kleśavad uktam ',
 "prasaṃkhyāne 'py akusīdasya sarvathā vivekakhyāter dharmameghaḥ samādhiḥ ",
 'tataḥ kleśakarmanivṛttiḥ ',
 'tadā sarvāvaraṇamalāpetasya jñānasyānantyāj jñeyam alpam ',
 'tataḥ kṛtārthānāṃ pariṇāmakramasamāptir guṇānām ',
 'kṣaṇapratiyogī pariṇāmāparāntanirgrāhyaḥ kramaḥ ',
 'puruṣārthaśūnyānāṃ guṇānāṃ pratiprasavaḥ kaivalyaṃ svarūpapratiṣṭhā vā citiśaktir iti ']


from process_sanskrit.utils.lexicalResources import filtered_words
from process_sanskrit.utils.dictionary_references import DICTIONARY_REFERENCES
from process_sanskrit.utils.lexicalResources import SANSKRIT_PREFIXES
from process_sanskrit.functions.dictionaryLookup import dict_search
import re
import regex

def extract_roots(list_of_entries, debug=False):
    roots = []
    i = 0
    
    while i < len(list_of_entries):
        current_entry = list_of_entries[i]
        
        # If entry has element at index 4 (original word)
        if len(current_entry) > 4:
            original_word = current_entry[4]
            stemmed_forms = []
            
            # Collect all stemmed forms for this original word
            j = i
            while j < len(list_of_entries) and len(list_of_entries[j]) > 4 and list_of_entries[j][4] == original_word:
                if list_of_entries[j][0] not in stemmed_forms:  # Avoid duplicates
                    stemmed_forms.append(list_of_entries[j][0])
                j += 1
            
            # Add as tuple if multiple stems, otherwise as single string
            if len(stemmed_forms) > 1:
                if debug:
                    print(f"Multiple stems for '{original_word}': {stemmed_forms}")
                roots.append(tuple(stemmed_forms))
            else:
                if debug:
                    print(stemmed_forms[0])
                roots.append(stemmed_forms[0])
                
            i = j  # Skip already processed entries
        else:
            # Handle entries without entry[4]
            if not roots or (isinstance(roots[-1], str) and roots[-1] != current_entry[0]) or \
               (isinstance(roots[-1], tuple) and current_entry[0] not in roots[-1]):
                if debug:
                    print(current_entry[0])
                roots.append(current_entry[0])
            i += 1
    
    return roots

def roots_splitted(list_of_entries, debug=False):

        root_dict = {}
        separators = r"[-—,/]"
        for entry in list_of_entries:
            if len(entry) == 7:
                parts = re.split(separators, entry[5])
                parts = [regex.sub(r'[^\p{L}]', '', part) for part in parts if part]
                parts = list(dict.fromkeys(parts))  # Remove duplicates while preserving order
                if entry[0] not in root_dict:
                    root_dict[entry[0]] = parts
            elif len(entry) == 3:
                parts = re.split(separators, entry[1])
                parts = [regex.sub(r'[^\p{L}]', '', part) for part in parts if part]
                parts = list(dict.fromkeys(parts))  # Remove duplicates while preserving order
                if entry[0] not in root_dict:
                    root_dict[entry[0]] = parts
        return root_dict
    

def clean_results(list_of_entries, mode="detailed", debug=False):

    i = 0
   
    #print("is it broken here?", list_of_entries)

    while i < len(list_of_entries) - 1:  # Subtract 1 to avoid index out of range error
        # Check if the word is in filtered_words
        if list_of_entries[i][0] in filtered_words:
            while i < len(list_of_entries) - 1 and list_of_entries[i + 1][0] == list_of_entries[i][0]:
                del list_of_entries[i + 1]

        if list_of_entries[i][0] == "api":
            if list_of_entries[i+1][0] == "apya" and list_of_entries[i+2][0] == "ap":
                del list_of_entries[i+1:i+2]
            if list_of_entries[i+1][0] == "ap":
                del list_of_entries[i+1]

        ### make a new rule for api ---> if word is api and the next entries are apya and ap remove them directly

        if list_of_entries[i][0] == "duḥ" and list_of_entries[i+1][0] == "kha":
            replacement = dict_search(["duḥkha"])
            if replacement is not None:
                list_of_entries[i] = replacement[0]
                del list_of_entries[i + 1]
                if list_of_entries[1+2] == "kha":
                    del list_of_entries[i + 2]  ##it's kha also as well

        if len(list_of_entries[i]) >= 5 and list_of_entries[i][0][-1] == "n" and list_of_entries[i][4] != list_of_entries[i][0]:
            #print("the one not replaced:", list_of_entries[i])
            if list_of_entries[i][4] in DICTIONARY_REFERENCES.keys():
                replacement = dict_search([list_of_entries[i][4]])
                if replacement is not None:
                    list_of_entries[i] = replacement[0]
        

        
        # Check if the word is "sam"
        if list_of_entries[i][0] == "sam":
            j = i + 1
            while j < len(list_of_entries) and (list_of_entries[j][0] == "sa" or list_of_entries[j][0] == "sam"):
                j += 1
            if j < len(list_of_entries):


                                ##non ha senso CHECK IF sam or sam  + list_of_entries[j][0]] are in MW dict
                                ## a quel punto fai voc_entry di quello, e rimpiazza tutte le entry inutili.
                
                voc_entry = None
                if list_of_entries[j][0] not in SANSKRIT_PREFIXES:
                    voc_entry = dict_search(["sam" + list_of_entries[j][0]])
                #print("voc_entry", voc_entry)

                ##non ha senso
                ## TODO replace this entirely
                
                # Ensure voc_entry is not None and has the expected structure
                if (voc_entry is not None and len(voc_entry) > 0 and 
                    isinstance(voc_entry[0], list) and len(voc_entry[0]) > 2 and 
                    isinstance(voc_entry[0][2], dict) and 'MW' in voc_entry[0][2]):
                    
                    # Check if the first key of the dictionary inside MW matches the condition
                    first_key = next(iter(voc_entry[0][2]['MW']), None)
                    if first_key and voc_entry[0][0] == first_key:
                        #print("revised query", ["saṃ" + list_of_entries[j][0]])
                        voc_entry = dict_search("saṃ" + list_of_entries[j][0])
                        #print("revise_voc_entry", voc_entry)
        
                if voc_entry is not None:
                    list_of_entries[i] = [item for sublist in voc_entry for item in sublist]
                    del list_of_entries[i + 1:j + 1]
        
        # Check if the word is "anu"
        if list_of_entries[i][0] == "anu":
            j = i + 1
            while j < len(list_of_entries) and (list_of_entries[j][0] == "anu"):
                j += 1
            if j < len(list_of_entries):
                voc_entry = dict_search(["anu" + list_of_entries[j][0]])
                ## testing to see if the check works. 
                print(voc_entry)
                if voc_entry[0][0] != voc_entry[0][2][0]:
                    list_of_entries[i] = [item for sublist in voc_entry for item in sublist]
                    del list_of_entries[i + 1:j + 1]
        
        # Check if the word is "ava"
        if list_of_entries[i][0] == "ava":
            j = i + 1
            while j < len(list_of_entries) and (list_of_entries[j][0] == "ava"):
                j += 1
            if j < len(list_of_entries):
                #print("testing with:", ["ava" + list_of_entries[j + 1][0]])
                
                voc_entry = dict_search(["ava" + list_of_entries[j + 1][0]])
                if voc_entry[0][0] != voc_entry[0][2][0]:
                    list_of_entries[i] = [item for sublist in voc_entry for item in sublist]
                    del list_of_entries[i + 1:j + 1]        
        i += 1  
    

    if mode == "parts":
            return roots_splitted(list_of_entries, debug=debug)
    elif mode == "roots":
            return extract_roots(list_of_entries, debug=debug)
    else:  # Default case when roots is "none" or any other value
        return list_of_entries



"""
Database setup and connection management for Sanskrit processing.

This module provides optimized database access patterns for:
1. Local text processing with many queries per word
2. Web application usage with concurrent requests

Features:
- Lazy loading to avoid unnecessary database connections
- Connection pooling for efficient query execution
- Session management utilities for reusing connections
- Proper path resolution for database file
- Robust error handling for missing database scenarios
"""

import os
import importlib.resources
import logging
import threading
from functools import wraps
from contextlib import contextmanager

from typing import Optional, Callable, Any, Generator, TypeVar, cast

from sqlalchemy import create_engine, Column, String, event
from sqlalchemy.exc import OperationalError
from sqlalchemy.engine import Engine
from sqlalchemy.orm import sessionmaker, scoped_session, declarative_base, Session

# Configure logging
log = logging.getLogger(__name__)

# Type variable for decorator return type preservation
F = TypeVar('F', bound=Callable[..., Any])

# --- Custom Exceptions ---
class DatabaseNotFoundError(Exception):
    """Raised when the required database file is not found."""
    pass

# --- Base Model Definition ---
Base = declarative_base()

# --- Global Variables for Lazy Loading ---
_engine: Optional[Engine] = None
_session_factory: Optional[sessionmaker] = None
_scoped_session: Optional[scoped_session] = None
_db_path_checked: bool = False

# --- Database Path Resolution ---
def get_db_path() -> str:
    """
    Get the path to the SQLite database file using importlib.resources.
    
    Returns:
        str: The resolved path to the database file
        
    Note:
        Uses a fallback path if importlib.resources resolution fails
    """
    try:
        # Modern approach (Python 3.9+)
        resources_dir = importlib.resources.files('process_sanskrit').joinpath('resources')
        db_file = resources_dir.joinpath('SQliteDB.sqlite')
        
        # Convert to string path
        with importlib.resources.as_file(db_file) as path:
            return str(path)
    except (ImportError, ModuleNotFoundError, AttributeError, FileNotFoundError) as e:
        # Fallback for older Python versions or unexpected package structure
        log.warning(f"Could not resolve database path with importlib: {e}")
        
        # Try relative resolution from this file
        module_dir = os.path.dirname(os.path.abspath(__file__))
        parent_dir = os.path.dirname(module_dir)  # Up one level from utils
        fallback_path = os.path.join(parent_dir, 'resources', 'SQliteDB.sqlite')
        
        log.info(f"Using fallback database path: {fallback_path}")
        return fallback_path

def database_exists(db_path: Optional[str] = None) -> bool:
    """
    Check if the database file exists at the expected path.
    
    Args:
        db_path: Optional explicit path to check
        
    Returns:
        bool: True if database exists, False otherwise
    """
    if db_path is None:
        db_path = get_db_path()
    return os.path.exists(db_path)

# --- Engine and Session Management ---
def get_engine(db_path: Optional[str] = None) -> Engine:
    """
    Lazily create and return the SQLAlchemy engine with optimized pooling.
    
    Args:
        db_path: Optional explicit database path
        
    Returns:
        Engine: Configured SQLAlchemy engine
        
    Raises:
        DatabaseNotFoundError: If database file doesn't exist
    """
    global _engine, _db_path_checked
    
    if _engine is None:
        if db_path is None:
            db_path = get_db_path()
        
        # Check database existence
        if not _db_path_checked and not database_exists(db_path):
            _db_path_checked = True
            error_msg = (
                f"Database file not found at: {db_path}\n"
                "Please run 'update-ps-database' to download and setup the database."
            )
            log.error(error_msg)
            raise DatabaseNotFoundError(error_msg)
        
        _db_path_checked = True
        DATABASE_URL = f"sqlite:///{db_path}"
        
        try:
            # Create engine with optimized pooling configuration for Sanskrit processing
            _engine = create_engine(
                DATABASE_URL,
                pool_size=5,          # Keep 5 connections ready
                max_overflow=10,      # Allow up to 10 more during peak loads
                pool_timeout=30,      # Wait up to 30 seconds for a connection
                pool_recycle=1800,    # Recycle connections after 30 minutes
                connect_args={"check_same_thread": False}  # Allow threads to share connections
            )
            
            # Configure SQLite for better performance and concurrency
            @event.listens_for(_engine, "connect")
            def set_sqlite_pragma(dbapi_connection, connection_record):
                cursor = dbapi_connection.cursor()
                try:
                    cursor.execute("PRAGMA journal_mode=WAL;")  # Write-Ahead Logging for better concurrency
                    cursor.execute("PRAGMA busy_timeout=5000;")  # Wait up to 5s if db is locked
                    cursor.execute("PRAGMA synchronous=NORMAL;")  # Faster with slight durability tradeoff
                    cursor.execute("PRAGMA foreign_keys=ON;")    # Enforce foreign key constraints
                    cursor.execute("PRAGMA cache_size=-64000;")  # Use ~64MB memory cache (negative = kilobytes)
                except Exception as e:
                    log.warning(f"Could not set all PRAGMA settings: {e}")
                finally:
                    cursor.close()
                    
            # Test connection
            with _engine.connect() as conn:
                log.info(f"Successfully connected to database at {db_path}")
                
        except OperationalError as e:
            _engine = None
            raise DatabaseNotFoundError(f"Failed to connect to database: {e}") from e
        except Exception as e:
            _engine = None
            log.error(f"Unexpected error creating engine: {e}")
            raise
            
    return _engine

def get_session_factory() -> sessionmaker:
    """
    Lazily create and return the SQLAlchemy session factory.
    
    Returns:
        sessionmaker: Configured session factory bound to the engine
    """
    global _session_factory
    
    if _session_factory is None:
        engine = get_engine()
        _session_factory = sessionmaker(bind=engine)
        log.debug("Created new session factory")
        
    return _session_factory

def get_scoped_session() -> scoped_session:
    """
    Get a thread-local scoped session for multi-threaded web applications.
    
    Returns:
        scoped_session: Thread-local session factory
        
    Note:
        Use this in web applications for thread safety
    """
    global _scoped_session
    
    if _scoped_session is None:
        session_factory = get_session_factory()
        _scoped_session = scoped_session(
            session_factory,
            scopefunc=threading.current_thread
        )
        log.debug("Created new scoped session")
        
    return _scoped_session

def get_session() -> Session:
    """
    Get a new SQLAlchemy session.
    
    Returns:
        Session: A new database session
        
    Note:
        For most operations, consider using session_scope() instead
    """
    session_factory = get_session_factory()
    return session_factory()

# --- Session Management Utilities ---
@contextmanager
def session_scope() -> Generator[Session, None, None]:
    """
    Context manager that handles session lifecycle and transactions.
    
    Yields:
        Session: A session that will be automatically committed or
                rolled back based on whether an exception occurs
                
    Usage:
        with session_scope() as session:
            results = session.query(Model).all()
    """
    session = get_session()
    try:
        yield session
        session.commit()
    except Exception:
        session.rollback()
        raise
    finally:
        session.close()

def with_session(func: F) -> F:
    """
    Decorator that provides a session to the decorated function.
    
    Args:
        func: Function that should receive a session parameter
        
    Returns:
        Wrapped function that automatically gets a session
        
    Usage:
        @with_session
        def process_word(word, session=None):
            # Use session for database operations
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        # Check if session already provided in kwargs
        if 'session' in kwargs and kwargs['session'] is not None:
            # Use existing session
            return func(*args, **kwargs)
        else:
            # Create new session
            with session_scope() as session:
                kwargs['session'] = session
                return func(*args, **kwargs)
    return cast(F, wrapper)

def requires_database(func: F) -> F:
    """
    Decorator that ensures database exists before calling function.
    
    Args:
        func: Function requiring database access
        
    Returns:
        Wrapped function that checks database existence
        
    Raises:
        DatabaseNotFoundError: If database file doesn't exist
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        if not database_exists():
            raise DatabaseNotFoundError(
                f"Function '{func.__name__}' requires database access, but database file not found. "
                "Please run 'update-ps-database' command to download the database."
            )
        return func(*args, **kwargs)
    return cast(F, wrapper)

# --- Model Definitions ---
class SplitCache(Base):
    """Table for caching sandhi split results to avoid redundant processing."""
    __tablename__ = 'split_cache'
    input = Column(String, primary_key=True)
    splitted_text = Column(String)

# --- Database Initialization ---
def initialize_database():
    """
    Create all defined tables in the database.
    
    Raises:
        DatabaseNotFoundError: If database file doesn't exist
    """
    try:
        engine = get_engine()
        log.info(f"Creating tables in database: {engine.url}")
        Base.metadata.create_all(engine)
        log.info("Tables created successfully")
    except DatabaseNotFoundError:
        log.error("Cannot initialize tables: database file not found")
        raise
    except Exception as e:
        log.error(f"Error during table creation: {e}")
        raise




from typing import List, Tuple, Dict, Union
from process_sanskrit.functions.enhancedSandhiSplitter import enhanced_sandhi_splitter
from process_sanskrit.functions.compoundAnalysis import root_compounds, process_root_result
from process_sanskrit.functions.sandhiSplitScorer import scorer


def hybrid_sandhi_splitter(
    text_to_split: str,
    cached: bool = False,
    attempts: int = 20,
    detailed_output: bool = False,
    score_threshold: float = 0.535,
    count_types: bool = False
) -> Union[List[str], Tuple[List[str], float, Dict, List], Tuple[List[str], Dict]]:
    """
    Enhanced sandhi splitter that combines statistical splitting with root compound analysis.
    Processes complex root analysis output into scoreable word lists.
    
    Parameters:
    - text_to_split: Text to split
    - cached: Whether to use caching
    - attempts: Number of splitting attempts for statistical method
    - detailed_output: If True, returns additional scoring information
    - score_threshold: Minimum score to accept statistical split
    """

     # Initialize counts dictionary if tracking
    compound_calls = 0 if count_types else None
    # First try our enhanced statistical splitter
    if detailed_output:
        stat_split, stat_score, stat_subscores, all_splits = enhanced_sandhi_splitter(
            text_to_split, cached, attempts, detailed_output=True
        )
        if len(stat_split) == 1:
            stat_score = 0 
    else:
        stat_split = enhanced_sandhi_splitter(
            text_to_split, cached, attempts, detailed_output=False
        )
        # Get the score for comparison
        if len(stat_split) == 1:
            stat_score = 0 
        else: 
            stat_score, stat_subscores = scorer.score_split(text_to_split, stat_split)


    # If score is good enough, return statistical result
    if stat_score >= score_threshold:
        if detailed_output:
            print("stat_score", stat_score)
            print("stat_split", stat_split)
        if count_types:
            return compound_calls
        if detailed_output:
            return stat_split, stat_score, stat_subscores, all_splits
        return stat_split

    # If score is too low, try root compound analysis
    try:
        if count_types:
            compound_calls += 1
        if detailed_output == True:
            print("text_to_split", text_to_split)
        root_analysis = root_compounds(text_to_split, inflection=False)
        if detailed_output == True:
            print("root_analysis", root_analysis)
        if root_analysis:
            # Process the root analysis results into a simple word list
            root_split = [process_root_result(item) for item in root_analysis]
            root_split = [x for i, x in enumerate(root_split) if i == 0 or x != root_split[i-1]]
            if detailed_output == True:
                print("root_split", root_split)
            # Score the processed root split
            root_score, root_subscores = scorer.score_split(text_to_split, root_split)
            if detailed_output == True:
                print("root_score", root_score)
                print("root_subscores", root_subscores)

            # Compare scores and choose the better result
            if root_score > stat_score:
                if count_types:
                    return compound_calls
                if detailed_output:
                    # Add root split to all_splits for reference
                    all_splits = [(root_split, root_score, root_subscores)] + (all_splits if all_splits else [])
                    return root_split, root_score, root_subscores, all_splits
                return root_split

    except Exception as e:
        print(f"Root compound analysis failed: {str(e)}")

    if count_types:
        return compound_calls

    # Fall back to statistical split if root analysis fails or scores lower
    if detailed_output:
        print("stat_split", stat_split)
        return stat_split, stat_score, stat_subscores, all_splits

    return stat_split





# Import main utility functions to expose at the module level
from process_sanskrit.utils.transliterationUtils import transliterate
from process_sanskrit.utils.detectTransliteration import detect
from process_sanskrit.utils.loadResources import get_resource_path
